use redis::{Client, Commands, Connection};
use ron::de::from_str;
use crate::device::Status;
use crate::errors::{LJError, LJResult};
use crate::worldstate::{WorldState, EDH};
// use log::info;

#[repr(u8)]
#[derive(Debug, PartialEq)]
pub enum Order {
	Draw = 0,
	Edh,
	Black,
	Grid,
	Resampler,
	ClientKey,
	Intensity,
	Kpps,
	ColorBalance,
	PowerOff
}

impl TryFrom<u8> for Order {
	type Error = String;

	fn try_from(value: u8) -> Result<Self, Self::Error> {
		use Order::*;

		if value > 8 {
			return Err("order out of range".to_string());
		}

		Ok(match value {
			0 => Draw,
			1 => Edh,
			2 => Black,
			3 => Grid,
			4 => Resampler,
			5 => ClientKey,
			6 => Intensity,
			7 => Kpps,
			8 => ColorBalance,
			9 => PowerOff,
			_ => unreachable!()
		})
	}
}

pub type Line = Vec<(f32, f32, u32)>;
pub type Resampler = Vec<Vec<(f32,f32)>>;

pub struct RedisCtrl {
	pub client: Client,
	pub connection: Connection,
	laser_id: u8,
}

impl RedisCtrl {
	pub fn new(url: &str, laser_id: &u8) -> LJResult<Self> {
		let client = Client::open(url)
			.map_err(LJError::RedisConnect)?;
		let connection = client.get_connection()
			.map_err(LJError::RedisConnect)?;
		Ok(RedisCtrl { client, connection, laser_id: *laser_id })
	}

	pub fn get_line(&mut self, key: &str) -> LJResult<Line> {
		let val: String = self.connection.get(key)?;
		let line: Line = from_str(&val)?;
		Ok(line)
	}

	pub fn set(&mut self, key: String, value: String) -> LJResult<()> {
		self.connection.set(key, value)?;
		Ok(())
	}

	pub fn get_order(&mut self, id: u8) -> LJResult<Order> {
		let path = format!("/order/{id}");
		let val: u8 = self.connection.get(path.clone())?;

		if val == 1 || val >= 4 {
			self.connection.set(path, 0)?;
		}

		Ok(val.try_into()?)
	}

	pub fn set_status(&mut self, status: Status) -> LJResult<()> {
		let lstt_key = format!("/lstt/{}", self.laser_id);
		let cap_key = format!("/cap/{}", self.laser_id);
		let lack_key = format!("/lack/{}", self.laser_id);
		self.set(lstt_key, status.playback_state.to_string())?;
		self.set(cap_key, status.capacity.to_string())?;
		self.set(lack_key, status.lack.to_string())?;
		Ok(())
	}

	pub fn init_world_state(&mut self) -> LJResult<WorldState> {
		Ok(WorldState {
			client_key: self.get_client_key()?,
			edh: self.get_edh()?,
			kpps: self.get_int("kpps")?.try_into()?,
			intensity: self.get_int("intensity")?.try_into()?,
			..WorldState::default()
		})
	}

	pub fn get_edh(&mut self) -> LJResult<EDH> {
		// Get new EDH
		let edh_key = format!("/EDH/{}", self.laser_id);
		let edh: String = self.connection.get(edh_key)?;
		let edh: Vec<Vec<f32>> = from_str(&edh)?;
		let edh = EDH::new(edh)?;
		Ok(edh)
	}

	pub fn get_client_key(&mut self) -> LJResult<String> {
		let key: String = self.connection.get("/clientkey")?;
		Ok(key)
	}

	pub fn get_color_balance(&mut self) -> LJResult<(u8, u8, u8)> {
		Ok((
			self.connection.get("/red")?,
			self.connection.get("/green")?,
			self.connection.get("/blue")?,
		))
	}

	pub fn get_resampler(&mut self ) -> LJResult<Resampler> {
		let val: String = self.connection.get(format!("/resampler/{}", self.laser_id))?;
		let resampler : Resampler = from_str(&val)?;
		Ok(resampler)
	}

	pub fn get_int(&mut self, key: &str) -> LJResult<u32> {
		// Get new Int
		let fmt = format!("/{key}/{}", self.laser_id);
		let val: u32 = self.connection.get(fmt)?;
		Ok(val)
	}
}
